import { KintoneFormFieldProperty, KintoneRestAPIClient } from '@kintone/rest-api-client';
import {
  AppID as KintoneAppId,
  Properties as KintoneFormFieldProperties,
} from '@kintone/rest-api-client/lib/src/client/types';
import { PluginConfigLookup } from './config';

export interface PluginContext {
  appId: KintoneAppId;
  formFieldsProperties: Record<KintoneAppId, KintoneFormFieldProperties>;
  attachmentFields: Record<KintoneAppId, KintoneFormFieldProperty.File[]>;
  lookupFields: KintoneFormFieldProperty.Lookup[];
}

const naturalCompare = (a: string, b: string): number => {
  const locales = new Set<string>([...navigator.languages, 'en-US', 'ja-JP']);
  return new Intl.Collator(Array.from(locales), {
    sensitivity: 'variant',
    numeric: true,
  }).compare(a, b);
};

const filterAttachmentFields = (properties: KintoneFormFieldProperties): KintoneFormFieldProperty.File[] => {
  return Object.values(properties)
    .filter((property) => property.type === 'FILE')
    .sort((a, b) => naturalCompare(`${a.label} (${a.code})`, `${b.label} (${b.code})`));
};

const filterLookupFields = (properties: KintoneFormFieldProperties): KintoneFormFieldProperty.Lookup[] => {
  return Object.values(properties)
    .filter((property) => 'lookup' in property)
    .sort((a, b) => naturalCompare(`${a.label} (${a.code})`, `${b.label} (${b.code})`));
};

const getFormFieldsProperties = async (appId: KintoneAppId): Promise<KintoneFormFieldProperties> => {
  const client = new KintoneRestAPIClient();
  const { properties } = await client.app.getFormFields({ app: appId, lang: 'en', preview: false });
  return properties;
};

export const getPluginContext = async (appId: KintoneAppId): Promise<PluginContext> => {
  const context: PluginContext = {
    appId: appId,
    formFieldsProperties: {},
    attachmentFields: {},
    lookupFields: [],
  };
  const properties = await getFormFieldsProperties(appId);
  context.formFieldsProperties[appId] = properties;
  context.attachmentFields[appId] = filterAttachmentFields(properties);
  const lookupFields = filterLookupFields(properties);
  const relatedAppIds = lookupFields
    .map(({ lookup }) => lookup.relatedApp.app)
    .filter((value, index, self) => self.indexOf(value) === index);
  await Promise.all(
    relatedAppIds.map(async (relatedAppId) => {
      const relatedProperties = await getFormFieldsProperties(relatedAppId);
      context.formFieldsProperties[relatedAppId] = relatedProperties;
      context.attachmentFields[relatedAppId] = filterAttachmentFields(relatedProperties);
    }),
  );
  context.lookupFields = lookupFields.filter((field) => {
    const relatedKeyField = context.formFieldsProperties[field.lookup.relatedApp.app][field.lookup.relatedKeyField];
    return (
      relatedKeyField.type === 'RECORD_NUMBER' ||
      ((relatedKeyField.type === 'SINGLE_LINE_TEXT' || relatedKeyField.type === 'NUMBER') &&
        'unique' in relatedKeyField &&
        relatedKeyField.unique)
    );
  });
  return context;
};

export const filterConfigByPluginContext = (
  config: PluginConfigLookup | null,
  context: PluginContext,
): PluginConfigLookup | null => {
  const filtered = config?.filter((item) => {
    const lookupField = context.lookupFields.find((f) => f.code === item.lookupFieldCode);
    return (
      lookupField != null &&
      context.attachmentFields[lookupField.lookup.relatedApp.app]?.find(
        (f) => f.code === item.srcAttachmentFieldCode,
      ) != null &&
      context.attachmentFields[context.appId]?.find((f) => f.code === item.destAttachmentFieldCode) != null
    );
  });
  return filtered != null && filtered.length > 0 ? filtered : null;
};
